﻿using Oracle.ManagedDataAccess.Client;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.Json;
using System.Text.RegularExpressions;
using System.Windows.Forms;

namespace DataGenerator
{
    public partial class MainForm : Form
    {
        private readonly Random _rnd = new(42);
        private const string TimestampPattern = "TO_TIMESTAMP('{0}', 'YYYY-MM-DD\"T\"HH24:MI')";
        private const string ConnectionString = @"Data Source=(DESCRIPTION=(ADDRESS_LIST=(ADDRESS=(PROTOCOL=TCP)(HOST=replace)(PORT=1521)))(CONNECT_DATA=(SERVICE_NAME=replace))));User Id=sys;DBA Privilege=SYSDBA;Password=password;Validate Connection=true";

        public MainForm()
        {
            InitializeComponent();
        }

        private void MainForm_Load(object sender, EventArgs e)
        {
            Directory.CreateDirectory("output");
        }

        private void btnGenerate_Click(object sender, EventArgs e)
        {
            var modelConfigString = Encoding.UTF8.GetString(Resource.ModelConfig);
            var modelConfig = JsonSerializer.Deserialize<ModelConfig>(modelConfigString);
            if (modelConfig == null)
            {
                throw new Exception("Model config not found");
            }

            foreach (var dimension in modelConfig.Dimensions)
            {
                GenerateBusinessDimension(dimension);
            }

            GenerateConsMecs(modelConfig.Mec);
            GenerateConsValues(modelConfig.Values);
        }

        private void btnInsert_Click(object sender, EventArgs e)
        {
            if (!Directory.Exists(txtInsertDirectory.Text))
            {
                lstOutput.Insert($"Directory '{txtInsertDirectory.Text}' not found");
                return;
            }

            if (ConnectionString.Contains("replace"))
            {
                lstOutput.Insert("Database connection string has placeholders");
                return;
            }

            var directories = Directory.GetDirectories(txtInsertDirectory.Text).OrderBy(s => s);

            var sw = new Stopwatch();
            sw.Start();

            foreach (var directory in directories)
            {
                ExecuteFiles(directory);
            }

            sw.Stop();
            lstOutput.Insert($"Inserted in {sw.Elapsed.TotalMilliseconds / 1000}s");
        }

        private void GenerateBusinessDimension(Dimension cfg)
        {
            var oldRows = new List<string>();
            var oldBusDims = new List<string>();
            var oldProjection = string.Join(", ", cfg.Old.Columns);

            var newRows = new List<string>();
            var newBusDims = new List<string>();
            var newProjection = string.Join(", ", cfg.New.Columns);

            for (var i = 1; i < cfg.Count + 1; i++)
            {
                var oldValues = new StringBuilder();
                var newValues = new StringBuilder();
                var index = i.ToString();

                // OID
                oldValues.Append(index);
                newValues.Append(index);

                // T103TIME_UNIT_VALUE
                if (cfg.Old.WithTu)
                {
                    oldValues.AppendWithComma(GetRandomOid());
                }

                if (cfg.New.WithTu)
                {
                    newValues.AppendWithComma(GetRandomOid());
                }

                // BD
                var bd = GetRandomOid();
                oldValues.AppendWithComma(bd);
                oldBusDims.Add(bd);
                newValues.AppendWithComma(index);
                newBusDims.Add(index);

                // VALIDITY
                oldValues.AppendWithComma(string.Format(TimestampPattern, "2015-01-01T22:00"));
                oldValues.AppendWithComma(string.Format(TimestampPattern, "2026-12-31T23:00"));
                newValues.AppendWithComma(string.Format(TimestampPattern, "2015-01-01T22:00"));
                newValues.AppendWithComma(string.Format(TimestampPattern, "2026-12-31T23:00"));

                // INTERNAL_ID
                oldValues.AppendWithComma(GetOldFullDimension($"EIC_{cfg.Name}{index}"));
                newValues.AppendWithComma(GetNewFullDimension($"EIC_{cfg.Name}{index}"));

                // NAME
                oldValues.AppendWithComma(GetOldFullDimension($"{cfg.Name}{index}"));
                newValues.AppendWithComma(GetNewFullDimension($"{cfg.Name}{index}"));

                if (cfg.Old.WithDfCode)
                {
                    // DF_CODE
                    oldValues.AppendWithComma(GetOldFullDimension($"{cfg.Name}_{index}"));
                    newValues.AppendWithComma(GetNewFullDimension($"{cfg.Name}_{index}"));
                }

                if (cfg.Old.WithCodingScheme)
                {
                    // CODING_SCHEME
                    oldValues.AppendWithComma(GetOldFullDimension(GetRandomOid()));
                    newValues.AppendWithComma(GetNewFullDimension(GetRandomInt(1, 6).ToString(CultureInfo.InvariantCulture)));
                }

                if (cfg.Old.WithShortName)
                {
                    // SHORT_NAME
                    oldValues.AppendWithComma(GetOldFullDimension($"{cfg.Name}{index}_S"));
                    newValues.AppendWithComma(GetNewFullDimension($"{cfg.Name}{index}_S"));
                }

                oldRows.Add($"INTO {cfg.Old.TableName} ({oldProjection}) VALUES ({string.Join(", ", oldValues)})");
                newRows.Add($"INTO {cfg.New.TableName} ({newProjection}) VALUES ({string.Join(", ", newValues)})");
            }

            oldRows.Insert(0, "INSERT ALL");
            oldRows.Add("SELECT 1 FROM DUAL;");
            oldRows.Add("COMMIT;");
            newRows.Insert(0, "INSERT ALL");
            newRows.Add("SELECT 1 FROM DUAL;");
            newRows.Add("COMMIT;");

            File.WriteAllLines(GetPath($"O_ROW_{cfg.Name}.sql"), oldRows);
            File.WriteAllLines(GetPath($"_O_BD_{cfg.Name}.txt"), oldBusDims);
            File.WriteAllLines(GetPath($"N_ROW_{cfg.Name}.sql"), newRows);
            File.WriteAllLines(GetPath($"_N_BD_{cfg.Name}.txt"), newBusDims);
        }

        private void GenerateConsMecs(Mec cfg)
        {
            var oldRows = new List<string>();
            var oldBusDim = new List<string>();
            var oldProjection = string.Join(", ", cfg.Old.Columns);
            var newRows = new List<string>();
            var newBusDim = new List<string>();
            var newProjection = string.Join(", ", cfg.New.Columns);

            var oldMga = File.ReadAllLines(GetPath("_O_BD_MGA.txt"));
            var oldRe = File.ReadAllLines(GetPath("_O_BD_RE.txt"));
            var oldCbt = File.ReadAllLines(GetPath("_O_BD_CONSUMPTION_TYPE.txt"));
            var oldMt = File.ReadAllLines(GetPath("_O_BD_MEASUREMENT_TYPE.txt"));
            var newMga = File.ReadAllLines(GetPath("_N_BD_MGA.txt"));
            var newRe = File.ReadAllLines(GetPath("_N_BD_RE.txt"));
            var newCbt = File.ReadAllLines(GetPath("_N_BD_CONSUMPTION_TYPE.txt"));
            var newMt = File.ReadAllLines(GetPath("_N_BD_MEASUREMENT_TYPE.txt"));

            var mecs = new List<MecConsRe>();

            for (var i = 1; i < cfg.Count + 1; i++)
            {
                MecConsRe mec;

                do
                {
                    mec = GetRandomMec(oldRe, oldMga, oldCbt, oldMt, newRe, newMga, newCbt, newMt);
                } while (mecs.Any(m => m.Equals(mec)));

                mecs.Add(mec);

                var oldValues = new StringBuilder();
                var newValues = new StringBuilder();
                var index = i.ToString();

                // OID
                oldValues.Append(index);
                newValues.Append(index);

                // T103TIME_UNIT_VALUE
                oldValues.AppendWithComma(GetRandomOid());

                // BD
                oldValues.AppendWithComma(mec.OldMga);
                oldValues.AppendWithComma(mec.OldRe);
                oldValues.AppendWithComma(mec.OldConsumptionType);
                oldValues.AppendWithComma(mec.OldMeasurementType);
                newValues.AppendWithComma(mec.NewMga);
                newValues.AppendWithComma(mec.NewRe);
                newValues.AppendWithComma(mec.NewConsumptionType);
                newValues.AppendWithComma(mec.NewMeasurementType);

                // VALIDITY
                oldValues.AppendWithComma(string.Format(TimestampPattern, "2015-01-01T22:00"));
                oldValues.AppendWithComma(string.Format(TimestampPattern, "2026-12-31T23:00"));
                newValues.AppendWithComma(string.Format(TimestampPattern, "2015-01-01T22:00"));
                newValues.AppendWithComma(string.Format(TimestampPattern, "2026-12-31T23:00"));

                // MEC_ID
                oldValues.AppendWithComma(GetOldFullDimension($"CONS_MEC{index}"));
                newValues.AppendWithComma(GetNewFullDimension($"CONS_MEC{index}"));

                oldRows.Add($"INTO {cfg.Old.TableName} ({oldProjection}) VALUES ({string.Join(", ", oldValues)})");
                oldBusDim.Add($"{mec.OldMga};{mec.OldRe};{mec.OldConsumptionType};{mec.OldMeasurementType}");
                newRows.Add($"INTO {cfg.New.TableName} ({newProjection}) VALUES ({string.Join(", ", newValues)})");
                newBusDim.Add($"{mec.NewMga};{mec.NewRe};{mec.NewConsumptionType};{mec.NewMeasurementType}");
            }

            oldRows.Insert(0, "INSERT ALL");
            oldRows.Add("SELECT 1 FROM DUAL;");
            oldRows.Add("COMMIT;");
            newRows.Insert(0, "INSERT ALL");
            newRows.Add("SELECT 1 FROM DUAL;");
            newRows.Add("COMMIT;");

            File.WriteAllLines(GetPath($"O_ROW_{cfg.Name}.sql"), oldRows);
            File.WriteAllLines(GetPath($"_O_BD_{cfg.Name}.txt"), oldBusDim);
            File.WriteAllLines(GetPath($"N_ROW_{cfg.Name}.sql"), newRows);
            File.WriteAllLines(GetPath($"_N_BD_{cfg.Name}.txt"), newBusDim);
        }

        private static void GenerateConsValues(Values cfg)
        {
            var oldRows = new List<string>();
            var oldMecs = File.ReadAllLines(GetPath($"_O_BD_{cfg.MecName}.txt"));
            var oldProjection = string.Join(", ", cfg.Old.Columns);

            var newRows = new List<string>();
            var newMecs = File.ReadAllLines(GetPath($"_N_BD_{cfg.MecName}.txt"));
            var newProjection = string.Join(", ", cfg.New.Columns);

            var day = new DateTime(2021, 1, 1);
            var oldIndex = 0;
            var newIndex = 0;

            var batch = cfg.Batch;
            var batches = Math.Ceiling(oldMecs.Length / (double) batch);

            for (var i = 0; i < cfg.CountOfDays; i++)
            {
                foreach (var mec in oldMecs)
                {
                    // BD
                    var bds = mec.Split(';');

                    var oldScript = $@"
                    INSERT INTO {cfg.Old.TableName} ({oldProjection})
                    SELECT
                        {oldIndex} + LEVEL,
                        14660000000000000 + LEVEL,
                        {bds[0]},
                        {bds[1]},
                        {bds[2]},
                        {bds[3]},
                        TO_TIMESTAMP('{day:yyyy-MM-dd}T00:00', 'YYYY-MM-DD""T""HH24:MI') + (1/{cfg.CountOfValuesInDay}) * (LEVEL - 1), 
                        TO_TIMESTAMP('{day:yyyy-MM-dd}T00:00', 'YYYY-MM-DD""T""HH24:MI') + (1/{cfg.CountOfValuesInDay}) * LEVEL,
                        0,
                        NULL,
                        TRUNC(DBMS_RANDOM.VALUE(100, 999), 8), 
                        'F', 
                        0, 
                        NULL, 
                        14660000000000000 + FLOOR(DBMS_RANDOM.VALUE(1, 5)), 
                        'F'
                    FROM DUAL
                    CONNECT BY LEVEL <= {cfg.CountOfValuesInDay};";

                    oldRows.Add(Regex.Replace(oldScript, @"\s+", " ").Trim());

                    oldIndex += cfg.CountOfValuesInDay;
                }

                foreach (var mec in newMecs)
                {
                    // BD
                    var bds = mec.Split(';');

                    var newScript = $@"
                    INSERT INTO {cfg.New.TableName} ({newProjection})
                    SELECT
                        {newIndex} + LEVEL,
                        {bds[0]},
                        {bds[1]},
                        {bds[2]},
                        {bds[3]},
                        TO_TIMESTAMP('{day:yyyy-MM-dd}T00:00', 'YYYY-MM-DD""T""HH24:MI') + (1/{cfg.CountOfValuesInDay}) * LEVEL,
                        TRUNC(DBMS_RANDOM.VALUE(100, 999), 7), 
                        'F',
                        FLOOR(DBMS_RANDOM.VALUE(1, 5)),
                        'F'
                    FROM DUAL
                    CONNECT BY LEVEL <= {cfg.CountOfValuesInDay};";

                    newRows.Add(Regex.Replace(newScript, @"\s+", " ").Trim());

                    newIndex += cfg.CountOfValuesInDay;
                }

                for (var j = 0; j < batches; j++)
                {
                    var oldRowsToWrite = oldRows
                        .Skip(j * batch)
                        .Take(batch);

                    var newRowsToWrite = newRows
                        .Skip(j * batch)
                        .Take(batch);

                    var date = $"{day:yyyy-MM-dd}";
                    Directory.CreateDirectory($@"output\O_{date}");
                    Directory.CreateDirectory($@"output\N_{date}");
                    File.WriteAllLines(GetPath($@"\O_{date}\O_GEN_{cfg.Name}_{date}_{j}.sql"), oldRowsToWrite);
                    File.WriteAllLines(GetPath($@"\N_{date}\N_GEN_{cfg.Name}_{date}_{j}.sql"), newRowsToWrite);
                }

                oldRows.Clear();
                newRows.Clear();

                day = day.AddDays(1);
            }
        }

        private void ExecuteFiles(string directory)
        {
            var files = Directory.GetFiles(directory).OrderBy(s => s);
            var sw = new Stopwatch();
            sw.Start();

            using var connection = new OracleConnection(ConnectionString);
            connection.Open();
            var command = connection.CreateCommand();

            foreach (var file in files)
            {
                var transaction = connection.BeginTransaction();

                var counter = 1;
                string line;

                var stream = new StreamReader(file);
                while ((line = stream.ReadLine()) != null)
                {
                    command.CommandText = line.Replace(";", "");

                    try
                    {
                        command.ExecuteNonQuery();
                    }
                    catch (Exception ex)
                    {
                        lstOutput.Insert(ex.Message);
                    }

                    counter++;
                }

                lstOutput.Insert($"File {file} is completed with {counter - 1} statements executed ({sw.Elapsed.TotalMilliseconds / 1000})");

                transaction.Commit();
            }
        }

        private MecConsRe GetRandomMec(string[] oldRes, string[] oldMgas, string[] oldCts, string[] oldMts,
            string[] newRes, string[] newMgas, string[] newCts, string[] newMts)
        {
            var rndRe = _rnd.Next(0, oldRes.Length);
            var rndMga = _rnd.Next(0, oldMgas.Length);
            var rndCt = _rnd.Next(0, oldCts.Length);
            var rndMt = _rnd.Next(0, oldMts.Length);

            var oldRe = oldRes[rndRe];
            var oldMga = oldMgas[rndMga];
            var oldCt = oldCts[rndCt];
            var oldMt = oldMts[rndMt];

            var newRe = newRes[rndRe];
            var newMga = newMgas[rndMga];
            var newCt = newCts[rndCt];
            var newMt = newMts[rndMt];

            return new MecConsRe
            {
                OldRe = oldRe,
                OldMga = oldMga,
                OldConsumptionType = oldCt,
                OldMeasurementType = oldMt,
                NewRe = newRe,
                NewMga = newMga,
                NewConsumptionType = newCt,
                NewMeasurementType = newMt
            };
        }

        private string GetRandomOid()
        {
            // OID has always length of 17
            return $"14660000{_rnd.Next(100000000, 999999999)}";
        }

        private double GetRandomInt(int min = 100, int max = 500)
        {
            return _rnd.Next(min, max);
        }

        private static string GetOldFullDimension(string tsValue)
        {
            // COMP_ERR, CORR_VALUE, TS_VALUE, T406VALSTATE
            return $"0, null, '{tsValue}', 'F'";
        }

        private static string GetNewFullDimension(string tsValue)
        {
            // VAL, VAL_S
            return $"'{tsValue}', 'F'";
        }

        private static string GetPath(string fileName)
        {
            return $@"{Directory.GetCurrentDirectory()}\output\{fileName}";
        }
    }
}